/*
 * hfscreenshot
 *
 * Writes the content of the Hifidelio display into a graphics format file.
 * 
 * v0.1 - First release with BMP output.
 * v0.5 - Added command line options and JPG support.
 */

#include <stdio.h>
#include <fcntl.h>
#include <getopt.h>
#include <setjmp.h>
#include "jpeglib.h"

// Display properties.

#define DEV_FB "/dev/fb0"
#define DISPLAY_WIDTH  400
#define DISPLAY_HEIGHT 160

// Image formats.

#define JPG_FORMAT 0
#define BMP_FORMAT 1
#define JPG_FILE_EXT ".jpg"
#define BMP_FILE_EXT ".bmp"

// BMP header; width: 400px, height: 160px, 1 bit per pixel, 2 bytes padding.

#define BMP_SIZE_HEADER  54
#define BMP_SIZE_PADDING 2
unsigned char bmp_header[] = {
	0x42, 0x4d, 0xbe, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3e, 0x00,
	0x00, 0x00, 0x28, 0x00, 0x00, 0x00, 0x90, 0x01, 0x00, 0x00, 0xa0, 0x00, 
	0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x20, 
	0x00, 0x00, 0x12, 0x0b, 0x00, 0x00, 0x12, 0x0b, 0x00, 0x00, 0x02, 0x00, 
	0x00, 0x00, 0x02, 0x00, 0x00, 0x00
};

// Screenshot colors (blue, green, red, reserved).

#define PALETTE_SIZE 8
#define COLOR_BLUE 0
#define COLOR_GREEN 1
#define COLOR_RED 2
unsigned char palette[] = { 0xcc, 0xcc, 0xaa, 0x00, 0x00, 0x00, 0x00, 0x00 };
unsigned char *back_color = palette;
unsigned char *fore_color = palette + 4;

// Default JPEG quality.

#define JPG_QUALITY 90

// Application name

char *app_name = NULL;


/* 
 * Error handling. Prints error message and exits program.
 */
void error(char *message) 
{
	fprintf(stderr, "%s: %s\n", app_name, message);
	exit(1);
}

/*
 * Dump display content into BMP file.
 */
void write_bmp(char *path) 
{
	int fd_fb;
	int fd_bmp;
	int y;
	int offset;
	int width_bytes = DISPLAY_WIDTH >> 3;
	unsigned char buffer[((DISPLAY_WIDTH >> 3) + BMP_SIZE_PADDING) * DISPLAY_HEIGHT];
	
	fd_fb = open(DEV_FB, O_RDONLY|O_NONBLOCK);
	if (fd_fb == -1) {
		error("error opening framebuffer device");
	}

	fd_bmp = open(path, O_CREAT|O_WRONLY, 0666);
	if (fd_bmp == -1) {
		error("error opening output file");
	}

	if (write(fd_bmp, bmp_header, BMP_SIZE_HEADER) == -1) {
		error("error writing image header");
	}
	if (write(fd_bmp, palette, PALETTE_SIZE) == -1) {
		error("error writing color palette");
	}
	for (y = DISPLAY_HEIGHT - 1; y >= 0; y--) {
		offset = y * (width_bytes + BMP_SIZE_PADDING);
		if (read(fd_fb, &(buffer[offset]), width_bytes) == -1) {
			error("error reading framebuffer");
		}
	}
	if (write(fd_bmp, &buffer, (width_bytes + BMP_SIZE_PADDING) * DISPLAY_HEIGHT) == -1) {
		error("error writing image data");
	}

	close(fd_fb);
	close(fd_bmp);
}	

/*
 * Dump display content into JPG file.
 */
void write_jpg(char *path, int quality) 
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	FILE *outfile;
	JSAMPROW row_pointer[1];
  	int row_stride;
	int fd_fb;
	int i, k = 0;
	unsigned char j;
	unsigned char display_byte[1];
	JSAMPLE image_buffer[3 * DISPLAY_WIDTH * DISPLAY_HEIGHT];

	fd_fb = open(DEV_FB, O_RDONLY|O_NONBLOCK);
	if (fd_fb == -1) {
		error("error opening framebuffer device");
	}

	for (i = 0; i < (DISPLAY_WIDTH >> 3) * DISPLAY_HEIGHT; i++) {
		if (read(fd_fb, &(display_byte), 1) == -1) {
			error("error reading framebuffer");
		}
		for (j = 0; j < 8; j++) {
			if ((*display_byte & (128 >> j)) == 0) {
				image_buffer[k++] = back_color[COLOR_RED];
				image_buffer[k++] = back_color[COLOR_GREEN];
				image_buffer[k++] = back_color[COLOR_BLUE];
			}
			else {
				image_buffer[k++] = fore_color[COLOR_RED];
				image_buffer[k++] = fore_color[COLOR_GREEN];
				image_buffer[k++] = fore_color[COLOR_BLUE];
			}
		}
	}
	
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);

	if ((outfile = fopen(path, "wb")) == NULL) {
		error("error opening output file");
	}
	jpeg_stdio_dest(&cinfo, outfile);

	cinfo.image_width = DISPLAY_WIDTH;
	cinfo.image_height = DISPLAY_HEIGHT;
	cinfo.input_components = 3;
	cinfo.in_color_space = JCS_RGB;
	
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE);
	jpeg_start_compress(&cinfo, TRUE);

	row_stride = DISPLAY_WIDTH * 3;

	while (cinfo.next_scanline < cinfo.image_height) {
    	row_pointer[0] = &image_buffer[cinfo.next_scanline * row_stride];
    	jpeg_write_scanlines(&cinfo, row_pointer, 1);
	}

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);
	
	fclose(outfile);
	close(fd_fb);
}

/*
 * Parse HTML color encoding
 */
int parse_color(char *color, unsigned char *parsed)
{
	int red, green, blue;

	sscanf(color, "%2x%2x%2x", &red, &green, &blue);
	parsed[COLOR_RED] = red;
	parsed[COLOR_GREEN] = green;
	parsed[COLOR_BLUE] = blue;
}

/* 
 * Print usage text.
 */
void usage() 
{
	printf("\nUsage: hfscreenshot [OPTION]... IMAGE-NAME\n");
	printf("\nOptions:");
	printf("\n  -i, --image-format  image format of the output file; possible values are 'jpg' or"); 
	printf("\n                      'bmp'; default is 'jpg'");
	printf("\n  -q, --jpeg-quality  jpeg quality range is 0 (worst) to 100 (best); default is 90");
	printf("\n  -f, --foreground    foreground color of the output image; default is hex '000000'");
	printf("\n  -b, --background    background color of the output image; default is hex 'aacccc'");
	printf("\n  -d, --target-dir    place output file in the given directory");
	printf("\n  -h, --help          display this help message\n\n");
	exit(0);
}

/*
 * Program entrance.
 */
int main(int argc, char **argv)
{
	int option_index = 0;
	int c;
	int illegal_option = FALSE;
	int image_format = JPG_FORMAT;
	int jpg_quality = JPG_QUALITY;
	char *target_dir = NULL;
	char *path = NULL;
	char *filename = NULL;
	
	static struct option long_options[] = {
		{"image-format", required_argument, 0, 'i'},
		{"jpeg-quality", required_argument, 0, 'q'},
		{"foreground",   required_argument, 0, 'f'},
		{"background",   required_argument, 0, 'b'},
		{"target-dir",   required_argument, 0, 'd'},
		{"help",         no_argument,       0, 'h'},
		{0, 0, 0, 0}
	};

	app_name = argv[0];
	
	while (!illegal_option && (c = getopt_long (argc, argv, "i:q:f:b:d:h", 
			long_options, &option_index)) != -1) {
		switch (c) {
			case 'i':
				if (!strcasecmp("bmp", optarg)) {
					image_format = BMP_FORMAT;
				}
				break;
			case 'q':
				jpg_quality = atoi(optarg);
				if (jpg_quality < 0 || jpg_quality > 100) {
					jpg_quality = JPG_QUALITY;
				}
				break;
			case 'f':
				parse_color(optarg, fore_color);
				break;
			case 'b':
				parse_color(optarg, back_color);
				break;
			case 'd':
				target_dir = (char *)malloc(strlen(optarg) + 1);
				strcpy(target_dir, optarg);
				break;
			case 'h':
				usage();
				break;
			case '?':
				illegal_option = TRUE;
				break;
			default:
				break;
		}
	}

	if (optind == argc) {
		fprintf(stderr, "%s: missing image name\n", argv[0]);
	}
	if (illegal_option || optind == argc) {
		fprintf(stderr, "Try 'hfscreenshot --help' for more information.\n");
	}
	else {
		filename = (char *)malloc(strlen(argv[optind]) 
			+ (image_format == JPG_FORMAT ? strlen(JPG_FILE_EXT) : strlen(JPG_FILE_EXT)) + 1);
		strcat(filename, argv[optind]);
		strcat(filename, (image_format == JPG_FORMAT ? JPG_FILE_EXT : BMP_FILE_EXT));
		
		if (target_dir != NULL) { 
			if (target_dir[strlen(target_dir) - 1] != '/') {
				target_dir = (char *)realloc(target_dir, strlen(target_dir) + 2);
				strcat(target_dir, "/");
			}
			path = (char *)malloc(strlen(target_dir) + strlen(filename) + 1);
			strcat (path, target_dir);
			strcat (path, filename);
		} 
		else {
			path = filename;
		}
				
		if (image_format == JPG_FORMAT) {
			write_jpg(path, jpg_quality);
		} 
		else {
			write_bmp(path);
		}
		printf("%s: file '%s' has been created\n", argv[0], path);
	}
}
